/*************************************************************************/
/* Depth upsampling from sparse data                                     */
/*                                                                       */
/* For the paper, please see:                                            */
/*  Dolson, J., Baek, J., Plagmann, C. and Thrun, S.,                    */
/*   ``Upsampling Range Data in Dynamic Environments,'' In CVPR 2010.    */
/*                                                                       */
/* This header implements an array that resides both on host and device, */
/* so that data transfer between the host machine and the GPU can        */
/* appear seamless.                                                      */
/*************************************************************************/

#ifndef MIRRORED_ARRAY_H
#define MIRRORED_ARRAY_H

#define CUDA_MEMORY_H

template<typename T> class MirroredArray {
 public:
  /* 
   * Constructor.
   *  len : the length of the array to be allocated.
   */
  MirroredArray(size_t len) {
    size = len;
    host = new T[len];
    owner = true;	
    allocateCudaMemory((void**)&(device), len*sizeof(T));
  }

  /*
   * Constructor with data provided.
   *  data : a pointer to the array residing on host.
   *  len  : the length of the array being provided. No
   *         additional memory on host will be allocated.
   */
  MirroredArray(T *data, size_t len) {
    size = len;
    host = data;
    owner = false;
    allocateCudaMemory((void**)&(device), len*sizeof(T));
    hostToDevice();
  }

  /*
   * Copies the host version of the data to the device.
   */
  void hostToDevice() {
    CUDA_SAFE_CALL(cudaMemcpy(device, host, size*sizeof(T), cudaMemcpyHostToDevice));
  }
    
  /*
   * Copies the device version of the data to the host.
   */
  void deviceToHost() {
    CUDA_SAFE_CALL(cudaMemcpy(host, device, size*sizeof(T), cudaMemcpyDeviceToHost));
  }

  /*
   * Destructor.
   */
  ~MirroredArray() {
    if (owner) delete[] host;
    releaseCudaMemory(device);
  }
  
  T *host;
  T *device;
  size_t size;
  bool owner;
};

#endif
