/*************************************************************************/
/* Depth upsampling from sparse data                                     */
/*                                                                       */
/* For the paper, please see:                                            */
/*  Dolson, J., Baek, J., Plagmann, C. and Thrun, S.,                    */
/*   ``Upsampling Range Data in Dynamic Environments,'' In CVPR 2010.    */
/*                                                                       */
/* This header provides wrappers around memory allocation/deallocation   */
/* routines in CUDA, for debugging purposes.                             */
/*************************************************************************/
#ifndef CUDA_MEMORY_H
#define CUDA_MEMORY_H

#include <stdio.h>

size_t GPU_MEMORY_ALLOCATION = 0;
size_t CUDA_MALLOC_COUNT = 0;

/*
 * Allocate memory on device.
 *  pointer : the return value for the memory allocated.
 *  size    : the size of the array to be allocated.
 */
void allocateCudaMemory(void ** pointer, size_t size) {
  CUDA_SAFE_CALL(cudaMalloc(pointer, size));
  CUDA_MALLOC_COUNT++;
  GPU_MEMORY_ALLOCATION += size;
}

/*
 * Release memory on device.
 *  pointer : the memory block to be released.
 */
void releaseCudaMemory(void * pointer) {
  CUDA_SAFE_CALL(cudaFree(pointer));
  CUDA_MALLOC_COUNT--;
}

/*
 * Report the number of unmatched allocation calls.
 */
void reportCudaMemory() {
  fprintf(stdout, "The # of outstanding cudaMalloc calls: %d\n",
	  (int)CUDA_MALLOC_COUNT);
}

/*
 * Reset the counter for allocation calls.
 */
void resetCudaMemoryUsage() {
  GPU_MEMORY_ALLOCATION = 0;
}

#endif
