/*************************************************************************/
/* Depth upsampling from sparse data                                     */
/*                                                                       */
/* For the paper, please see:                                            */
/*  Dolson, J., Baek, J., Plagmann, C. and Thrun, S.,                    */
/*   ``Upsampling Range Data in Dynamic Environments,'' In CVPR 2010.    */
/*                                                                       */
/* This is an adaptation of the permutohedral-lattice filtering          */
/* algorithm by Adams, A., Baek, J., and Davis, M. A.,                   */
/*  ``Fast High-Dimensional Filtering Using the Permutohedral Lattice,'' */
/*   In Eurographics 2010. to perform joint bilateral filtering on       */
/* sparse range data, given a number of image frames.                    */
/*                                                                       */
/*                                                                       */
/* Most routines here are helper functions. See filter_ for the wrapper. */
/*                                                                       */
/* All rights reserved.                                                  */
/*************************************************************************/

#define BLOCK_SIZE 64
#define _DEBUG
#include <cuda_runtime.h>
#include <stdio.h>

#include "cutil.h"
#include "cuda_memory.h"
#include "MirroredArray.h"
#include "hash_table.cu"

#ifdef WIN32
#include "win32time.h"
#else
#include <sys/time.h>
#endif
  
#ifdef LIBRARY
extern "C"
#ifdef WIN32
__declspec(dllexport)
#endif
#endif

struct MatrixEntry {
    int index;
    float weight;
};

#define FIXED_BITS 12

/***************************************************/
/* Initialize CUDA and print a list of properties. */
/***************************************************/
void initCuda(int argc, char **argv) {
    CUT_DEVICE_INIT(argc, argv);    

    cudaDeviceProp prop;
    CUDA_SAFE_CALL_NO_SYNC(cudaGetDeviceProperties(&prop, 0));
    printf("Device name: %s\n", prop.name);
    printf("Max threads per block: %i\n", prop.maxThreadsPerBlock);
    printf("Max threads dim: %i %i %i\n", prop.maxThreadsDim[0],
	   prop.maxThreadsDim[1], prop.maxThreadsDim[2]);
    printf("Max grid size: %i %i %i \n", prop.maxGridSize[0],
	   prop.maxGridSize[1], prop.maxGridSize[2]);
    printf("Shared memory per block: %lu Kb\n", prop.sharedMemPerBlock/1024);
    printf("Total global memory: %lu Kb\n", prop.totalGlobalMem/1024);
    printf("Warp size: %i\n", prop.warpSize);
    printf("Memory pitch: %i\n", (int)prop.memPitch);
    printf("Registers per block: %i\n", prop.regsPerBlock);
    printf("Clock rate: %i\n", prop.clockRate);
    printf("Texture alignment: %i\n", (int)prop.textureAlignment);
    fflush(stdout);
}

/***************************************************/
/* Create the splatting and slicing matrix.        */
/***************************************************/
/* pd          : the number of dimensions in position vectors
 * n           : the number of position vectors
 * positions   : a pointer to the array of position vectors (n * pd floats)
 * scaleFactor : a scalar to be applied to the position vectors
 * matrix      : an array of matrix entries (allocated already)
 * table       : the hash table to be populated
 * zeros       : an output array that will contain the zero-color neighbors of each vector.
 * ranks       : an output array that will contain the rank of the offset vectors for each vector.
 *
 * The last two arguments are used only if LINEAR_D_MEMORY is on.
 */
template<int pd>
__global__ static void createMatrix(const int n,
				    const float *positions, 
				    const float *values, 
				    const float *scaleFactor,
				    MatrixEntry *matrix, HashTable *table,
				    signed short  *zeros = NULL,
				    char *ranks = NULL) {
  const int idx = threadIdx.x + blockIdx.x * blockDim.x;
  const int threadId = threadIdx.y*blockDim.x + threadIdx.x;
  const bool outOfBounds = (idx >= n);
  
  float myElevated[pd+1];
  const float *myPosition = positions + idx*pd;
  
#ifdef LINEAR_D_MEMORY
  signed short * myGreedy = zeros + idx*(pd+1);
  char * myRank = ranks + idx*(pd+1);
#else
  int myGreedy[pd+1];
  int myRank[pd+1];
#endif
  
  float myBarycentric[pd+2];
  __shared__ short keys[pd*BLOCK_SIZE];
  short *myKey = keys + threadId * pd;
  __shared__ int indices[BLOCK_SIZE];
  
  if (!outOfBounds) {
    // Elevate the point vector to the (d+1)-dimensional hyperplane.
    myElevated[pd] = -pd*(myPosition[pd-1])*scaleFactor[pd-1];
    for (int i = pd-1; i > 0; i--) {
      myElevated[i] = (myElevated[i+1] - 
		       i*(myPosition[i-1])*scaleFactor[i-1] + 
		       (i+2)*(myPosition[i])*scaleFactor[i]);
    }
    myElevated[0] = myElevated[1] + 2*(myPosition[0])*scaleFactor[0];
    
    // greedily search for the closest zero-colored lattice point.
    signed short sum = 0;
    for (int i = 0; i <= pd; i++) {
      float v = myElevated[i]*(1.0f/(pd+1));
      float up = ceilf(v) * (pd+1);
      float down = floorf(v) * (pd+1);
      if (up - myElevated[i] < myElevated[i] - down) {
	myGreedy[i] = (signed short)up;
      } else {
	myGreedy[i] = (signed short)down;
      }
      sum += myGreedy[i];
    }
    sum /= pd+1;
    
    // sort differential to find the permutation between the simplex containing the point vector
    // and the canonical one
    for (int i = 0; i <= pd; i++) {
      myRank[i] = 0;
      for (int j = 0; j <= pd; j++) {
	if (myElevated[i] - myGreedy[i] < myElevated[j] - myGreedy[j] ||
	    (myElevated[i] - myGreedy[i] == myElevated[j] - myGreedy[j]
	     && i > j)) {
	  myRank[i]++;
	}
      }
    }	
    if (sum > 0) { // sum too large, need to bring down the ones with the smallest differential
      for (int i = 0; i <= pd; i++) {
	if (myRank[i] >= pd + 1 - sum) {
	  myGreedy[i] -= pd+1;
	  myRank[i] += sum - (pd+1);
	} else {
	  myRank[i] += sum;
	}
      }
    } else if (sum < 0) { // sum too small, need to bring up the ones with largest differential
      for (int i = 0; i <= pd; i++) {
	if (myRank[i] < -sum) {
	  myGreedy[i] += pd+1;
	  myRank[i] += (pd+1) + sum;
	} else {
	  myRank[i] += sum;
	}
      }
    }
    
    // turn delta into barycentric coordinates
    for (int i = 0; i <= pd+1; i++) {
      myBarycentric[i] = 0;
    }      
    for (int i = 0; i <= pd; i++) {
      float delta = (myElevated[i] - myGreedy[i]) * (1.0f/(pd+1));
      myBarycentric[pd-myRank[i]] += delta;
      myBarycentric[pd+1-myRank[i]] -= delta;
    }
    myBarycentric[0] += 1.0f + myBarycentric[pd+1];
  }
  
#ifdef USE_ADDITIVE_HASH
  unsigned int cumulative_hash = hash<pd>(table, myGreedy);
#endif
  for (int color = 0; color <= pd; color++) {
    // Compute the location of the lattice point explicitly (all but
    // the last coordinate - it's redundant because they sum to zero)
    if (!outOfBounds) {
      for (int i = 0; i < pd; i++) {
	myKey[i] = myGreedy[i] + color;
	if (myRank[i] > pd-color) myKey[i] -= (pd+1);
      }
    }
    
    __syncthreads();
    
    // Find the lowest numbered thread interested in this key
    int lowestThread;   
    if (!outOfBounds) {
      for (lowestThread = 0; lowestThread < threadId; lowestThread++) {
	bool match = true;
	for (int j = 0; j < pd && match; j++) {
	  match = (myKey[j] == keys[lowestThread*pd+j]);
	}
	if (match) break;
      }
      
      // If I'm the lowest numbered interested thread, do the hash table lookup
      if (lowestThread == threadId) {
	indices[lowestThread] = hashTableInsert<pd>(table, myKey, idx*(pd+1)+color);
      }
    }
#ifdef USE_ADDITIVE_HASH
    for (int i = 0; i < pd; i++) { if (myRank[i] == pd-color) cumulative_hash += hOffset[i]; }
#endif
    
    __syncthreads();
    
    if (!outOfBounds) {	    
      // Record this matrix entry
      MatrixEntry r;
      r.index = indices[lowestThread];
      r.weight = myBarycentric[color];
      matrix[idx*(pd+1) + color] = r;
    }
  }    
}

/***************************************************/
/* Clean the hash table after initialization. This */
/* is necessary for removing duplicates.           */
/***************************************************/
template<int kd>
__global__ static void cleanHashTable(int n, HashTable *table, MatrixEntry *matrix) {
  const int idx = (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x * blockDim.y + threadIdx.x;
  
  if (idx >= n) return;
  
  // find my hash table entry
  int *e = table->entries + idx;
  
  // Check if I created my own key in the previous phase
  if (*e >= 0) {
    // Rehash my key and reset the pointer in order to merge with
    // any other pixel that created a different entry under the
    // same key. If the computation was serial this would never
    // happen, but sometimes race conditions can make the same key
    // be inserted twice. hashTableRetrieve always returns the
    // earlier, so it's no problem as long as we rehash now.
    
#ifdef LINEAR_D_MEMORY
    // Get my key      
    short myKey[kd];
    generateKey<kd>(table->zero, table->rank, *e, myKey);
    *e = hashTableRetrieve<kd>(table, myKey);
#else
    *e = hashTableRetrieve<kd>(table, table->keys + *e*kd);
#endif
  }
}

/***************************************************/
/* Splat the values into the hash table.           */
/***************************************************/
template<int pd, int vd>
  __global__ static void splatCache(const int n, HashTable *table, float *values, MatrixEntry *matrix) {
  
  const int idx = threadIdx.x + blockIdx.x * blockDim.x;
  const int threadId = threadIdx.y*blockDim.x + threadIdx.x;
  const int color = blockIdx.y % (pd+1);
  const bool outOfBounds = (idx >= n);

  __shared__ int sharedOffsets[BLOCK_SIZE];
  __shared__ unsigned int sharedValues[BLOCK_SIZE*(vd+1)];
  int myOffset = -1;
  unsigned int *myValue = sharedValues + threadId*(vd+1);
    
  if (!outOfBounds) {    
    float *value = values + idx*vd;    
    MatrixEntry r = matrix[idx*(pd+1)+color];

    // convert the matrix entry from a pointer into the entries array to a pointer into the keys/values array
    matrix[idx*(pd+1)+color].index = r.index = table->entries[r.index];
    // record the offset into the keys/values array in shared space
    myOffset = sharedOffsets[threadId] = r.index*(vd+1);
    
    for (int j = 0; j < vd; j++) {
      myValue[j] = (unsigned int)(value[j]*r.weight*(1<<FIXED_BITS));
    }
    myValue[vd] = (unsigned int)(r.weight*(1<<FIXED_BITS));
    
  } else {
    sharedOffsets[threadId] = -1;
  }
  
  __syncthreads();    
  // am I the first thread in this block to care about this key?
  
  if (outOfBounds) return;
  
  for (int i = 0; i < BLOCK_SIZE; i++) { 
    if (i < threadId) {
      if (myOffset == sharedOffsets[i]) {
	// somebody else with higher priority cares about this key
	return;
      }
    } else if (i > threadId) {
      if (myOffset == sharedOffsets[i]) {
	// someone else with lower priority cares about this key, accumulate it into mine
	for (int j = 0; j <= vd; j++) {
	  sharedValues[threadId*(vd+1) + j] += sharedValues[i*(vd+1) + j];
	}
      }
    }
  }
    
  // only the threads with something to write to main memory are still going
  unsigned int *val = table->values + myOffset;
  for (int j = 0; j <= vd; j++) {
    atomicAdd(val+j, myValue[j]);
  }
}

/***************************************************/
/* Blur the lattice                                */
/***************************************************/
template<int pd, int vd>
  __global__ static void blur(int n, HashTable *table, unsigned int *newValues, MatrixEntry *matrix, int color) {
  const int idx = (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x * blockDim.y + threadIdx.x;
  
  if (idx >= n) return;
  
  // Check if I'm valid
  if (matrix[idx].index != idx) return;
  
  // find my key and the keys of my neighbours
  short myKey[pd+1];
  short np[pd+1];
  short nm[pd+1];
  
#ifdef LINEAR_D_MEMORY
  generateKey<pd>(table->zero, table->rank, idx, myKey);
  for (int i = 0; i < pd; i++) {
    np[i] = myKey[i]+1;    
    nm[i] = myKey[i]-1;
  }
#else
  for (int i = 0; i < pd; i++) {
    myKey[i] = table->keys[idx*pd+i];
    np[i] = myKey[i]+1;    
    nm[i] = myKey[i]-1;
  }
#endif
  
  np[color] -= pd+1;
  nm[color] += pd+1;
  
#ifdef USE_ADDITIVE_HASH
  unsigned int hCurrent = hash<pd>(table, myKey);
  int offNp = hashTableRetrieveWithHash<pd>(table, hCurrent+hOffset[color],np);
  int offNm = hashTableRetrieveWithHash<pd>(table, hCurrent-hOffset[color],nm);
#else
  int offNp = hashTableRetrieve<pd>(table, np);
  int offNm = hashTableRetrieve<pd>(table, nm);
#endif

  // Look up the values of the neighbors.
  unsigned int *valMe = table->values + (vd+1)*idx;
  unsigned int *valNp = table->values + (vd+1)*offNp;
  unsigned int *valNm = table->values + (vd+1)*offNm;	
  unsigned int *valOut = newValues + (vd+1)*idx;
  
  if (offNp >= 0 && offNm >= 0) {
    for (int i = 0; i <= vd; i++) {
      valOut[i] = (valNp[i] + (valMe[i]<<1) + valNm[i])>>2;
    }
  } else if (offNp >= 0) {
    for (int i = 0; i <= vd; i++) {
      valOut[i] = (valNp[i] + (valMe[i]<<1))>>2;
    }
  } else if (offNm >= 0) {
    for (int i = 0; i <= vd; i++) {
      valOut[i] = (valNm[i] + (valMe[i]<<1))>>2;
    }
  } else {
    for (int i = 0; i <= vd; i++) {
      valOut[i] = valMe[i]>>1;	    
    }
  }    
}

/***************************************************/
/* Slice out of the lattice                        */
/***************************************************/
/*
 * m           : the number of position vectors to sample
 * positions   : the position vectors (m * pd floats)
 * values      : an output array that will contain the sliced values
 * scaleFactor : a scalar afctor on the positions
 * matrix      : matrix containing the lattce points
 * hash table  : the hash table containing the position vectors as keys
 * zeros, ranks: use these (created earlier) if LINEAR_D_MEMORY is set.
 */
template<int pd, int vd> __global__
  static void readMatrix(const int m,
			 const float *positions, 
			 float *values, 
			 const float *scaleFactor,
			 MatrixEntry *matrix, HashTable *table,
			 signed short  *zeros = NULL,
			 char *ranks = NULL) {  
  const int idx = threadIdx.x + blockIdx.x * blockDim.x;
  const int threadId = threadIdx.y*blockDim.x + threadIdx.x;
  const bool outOfBounds = (idx >= m);
  
  float myElevated[pd+1];
  const float *myPosition = positions + idx*pd;
  
#ifdef LINEAR_D_MEMORY
  signed short * myGreedy = zeros + idx*(pd+1);
  char * myRank = ranks + idx*(pd+1);
#else
  int myGreedy[pd+1];
  int myRank[pd+1];
#endif
  
  float myBarycentric[pd+2];
  __shared__ short keys[pd*BLOCK_SIZE];
  short *myKey = keys + threadId * pd;
  
  if (!outOfBounds) {
    
    myElevated[pd] = -pd*(myPosition[pd-1])*scaleFactor[pd-1];
    for (int i = pd-1; i > 0; i--) {
      myElevated[i] = (myElevated[i+1] - 
		       i*(myPosition[i-1])*scaleFactor[i-1] + 
		       (i+2)*(myPosition[i])*scaleFactor[i]);
    }
    myElevated[0] = myElevated[1] + 2*(myPosition[0])*scaleFactor[0];
    
    // find the closest zero-colored lattice point
    
    // greedily search for the closest zero-colored lattice point
    signed short sum = 0;
    for (int i = 0; i <= pd; i++) {
      float v = myElevated[i]*(1.0f/(pd+1));
      float up = ceilf(v) * (pd+1);
      float down = floorf(v) * (pd+1);
      if (up - myElevated[i] < myElevated[i] - down) {
	myGreedy[i] = (signed short)up;
      } else {
	myGreedy[i] = (signed short)down;
      }
      sum += myGreedy[i];
    }
    sum /= pd+1;
    
    // sort differential to find the permutation between this simplex and the canonical one
    for (int i = 0; i <= pd; i++) {
      myRank[i] = 0;
      for (int j = 0; j <= pd; j++) {
	if (myElevated[i] - myGreedy[i] < myElevated[j] - myGreedy[j] ||
	    (myElevated[i] - myGreedy[i] == myElevated[j] - myGreedy[j]
	     && i > j)) {
	  myRank[i]++;
	}
      }
    }
    
    if (sum > 0) { // sum too large, need to bring down the ones with the smallest differential
      for (int i = 0; i <= pd; i++) {
	if (myRank[i] >= pd + 1 - sum) {
	  myGreedy[i] -= pd+1;
	  myRank[i] += sum - (pd+1);
	} else {
	  myRank[i] += sum;
	}
      }
    } else if (sum < 0) { // sum too small, need to bring up the ones with largest differential
      for (int i = 0; i <= pd; i++) {
	if (myRank[i] < -sum) {
	  myGreedy[i] += pd+1;
	  myRank[i] += (pd+1) + sum;
	} else {
	  myRank[i] += sum;
	}
      }
    }
    
    // turn delta into barycentric coords
    for (int i = 0; i <= pd+1; i++) {
      myBarycentric[i] = 0;
    }
    
    for (int i = 0; i <= pd; i++) {
      float delta = (myElevated[i] - myGreedy[i]) * (1.0f/(pd+1));
      myBarycentric[pd-myRank[i]] += delta;
      myBarycentric[pd+1-myRank[i]] -= delta;
    }
    myBarycentric[0] += 1.0f + myBarycentric[pd+1];
  }
  
#ifdef USE_ADDITIVE_HASH
  unsigned int cumulative_hash = hash<pd>(table, myGreedy);
#endif
  float myweight = 0.f;
  float mysum[vd];
  for (int j = 0; j < vd; mysum[j++] = 0.f);
  if (!outOfBounds) {
    for (int color = 0; color <= pd; color++) {
      // Compute the location of the lattice point explicitly (all but
      // the last coordinate - it's redundant because they sum to zero)
      for (int i = 0; i < pd; i++) {
	myKey[i] = myGreedy[i] + color;
	if (myRank[i] > pd-color) myKey[i] -= (pd+1);
      }
      int idx = hashTableRetrieve<pd>(table, myKey);
      if (idx != -1) {
	for (int j = 0; j < vd; j++)
	  mysum[j] += table->values[idx*(vd+1)+j] * myBarycentric[color];
	myweight += table->values[idx*(vd+1)+vd] * myBarycentric[color];
	
      }
    }
    values[idx*(vd+1)+vd] = myweight;
    if (myweight > 0.f) {
      myweight = 1.f / myweight;
    }
    for (int j = 0; j < vd; j++)
      values[idx*(vd+1)+j] = mysum[j] * myweight;
  }  
}

/***************************************************/
/* A wrapper around other functions to perform     */
/* Gaussian filtering of the input data.           */
/***************************************************/
/*
 * im         : an array containing value vectors (size n * vd)
 * ref        : an array containing the corresponding position vectors (size n * pd)
 * sample     : an array containing position vectors to slice out (size m * pd)
 * output     : an output array that will contain the corresponding value vectors (size m * vd)
 * n          : the number of vectors in im and ref.
 * m          : the number of vectors in sample.
 * accurate   : a boolean flag that specifies whether the blur step is performed or not.
 */
template<int vd, int pd>
  void filter_(float *im, float *ref, float *sample, float * output, int n, int m, bool accurate) {    
  float blurVariance = accurate ? 0.5 : 0;
  
  MirroredArray<float> scaleFactor(pd);
  for (int i = 0; i < pd; i++)
    scaleFactor.host[i] = (pd+1)*sqrtf((1.0/6 + blurVariance)/((i+1)*(i+2)));
  scaleFactor.hostToDevice();
  
  MirroredArray<float> values(im, n*vd);
  MirroredArray<float> positions(ref, n*pd);
  MirroredArray<float> outputs(output, m*(vd+1));
  MirroredArray<float> samples(sample, m*pd);
  MirroredArray<MatrixEntry> matrix(n*(pd+1));
  HashTable *table = createHashTable<pd, vd+1>(n*(pd+1));
  
#ifdef LINEAR_D_MEMORY
  MirroredArray<char> ranks(n*(pd+1));
  MirroredArray<signed short> zeros(n*(pd+1));
  CUDA_SAFE_CALL(cudaMemcpy(&(table->zero), &(zeros.device), sizeof(signed short *), cudaMemcpyHostToDevice));
  CUDA_SAFE_CALL(cudaMemcpy(&(table->rank), &(ranks.device), sizeof(char *), cudaMemcpyHostToDevice));
#endif
  
  // Populate constant memory with hash of offset vectors
  unsigned int hOffset_host[pd+1];
  signed short offset[pd+1];
  for (int i = 0; i < pd; offset[i] = 1, i++);
  for (int i = 0; i <= pd; i++) {
    offset[i] -= pd+1; hOffset_host[i] = hash<pd>(table, offset); offset[i] += pd+1;
  }
  CUDA_SAFE_CALL(cudaMemcpyToSymbol((char*)&hOffset, &hOffset_host, sizeof(unsigned int)*(pd+1)));
  
  dim3 nblocks((n-1)/BLOCK_SIZE+1,1,1);
  dim3 blocksize(BLOCK_SIZE,1,1);
  
  timeval t[7];  
  /***************************************************/
  /* Create the lattice.                             */
  /***************************************************/
  gettimeofday(t+0, NULL);    
#ifdef LINEAR_D_MEMORY
  createMatrix<pd><<<nblocks, blocksize>>>(n, positions.device, values.device, scaleFactor.device,
					   matrix.device, table, zeros.device, ranks.device);
#else
  createMatrix<pd><<<nblocks, blocksize>>>(n, positions.device, values.device, scaleFactor.device,
					   matrix.device, table);
#endif
  CUT_CHECK_ERROR("Matrix creation failed\n");

  /***************************************************/
  /* Clean the hash table.                           */
  /***************************************************/
  gettimeofday(t+1, NULL);  
  // fix duplicate hash table entries
  int cleanBlockSize = 32;
  dim3 cleanBlocks((n-1)/cleanBlockSize+1, 2*(pd+1), 1);
  cleanHashTable<pd><<<cleanBlocks, cleanBlockSize>>>(2*n*(pd+1), table, matrix.device);
  CUT_CHECK_ERROR("clean failed\n");
  
  /***************************************************/
  /* Splat.                                          */
  /***************************************************/
  gettimeofday(t+2, NULL);    
  // splat splits by color, so extend the y coordinate to our blocks to represent that
  nblocks.y *= pd+1;
  splatCache<pd, vd><<<nblocks, blocksize>>>(n, table, values.device, matrix.device);
  CUT_CHECK_ERROR("splat failed\n");

  /***************************************************/
  /* Blur.                                           */
  /***************************************************/
  gettimeofday(t+3, NULL);        
  if (accurate) {
    unsigned int *newValues;
    allocateCudaMemory((void**)&(newValues), n*(pd+1)*(vd+1)*sizeof(unsigned int));
    CUDA_SAFE_CALL(cudaMemset((void *)newValues, 0, n*(pd+1)*(vd+1)*sizeof(unsigned int)));
    
    for (int color = 0; color <= pd; color++) {	
      blur<pd, vd><<<cleanBlocks, cleanBlockSize>>>(n*(pd+1), table, newValues, matrix.device, color);
      CUT_CHECK_ERROR("blur failed\n");
      newValues = swapHashTableValues(table, newValues);
    }
  }
  
  /***************************************************/
  /* Slice.                                          */
  /***************************************************/  
  gettimeofday(t+4, NULL);      
  nblocks.y /= (pd+1);
  int mblocks_max = 2<<14;
  int mdone = 0;
  while(mdone < m) {
    int mblocks_size = (m - mdone - 1) / BLOCK_SIZE + 1;
    mblocks_size = (mblocks_size > mblocks_max) ? mblocks_max : mblocks_size;
    int mleft = (m - mdone > mblocks_size * BLOCK_SIZE) ? mblocks_size * BLOCK_SIZE : m - mdone;
    dim3 mblocks(mblocks_size,1,1);
    readMatrix<pd,vd><<<mblocks, blocksize>>>(mleft, samples.device + mdone * pd,
					      outputs.device + mdone * (vd+1), scaleFactor.device,
#ifdef LINEAR_D_MEMORY
					      matrix.device, table, zeros.device, ranks.device
#else
					      matrix.device, table
#endif
					      );
    mdone += mleft;
  }
  CUT_CHECK_ERROR("slice failed\n");	

  /***************************************************/
  /* Finalize.                                       */
  /***************************************************/
  gettimeofday(t+5, NULL);      
  double total = (t[5].tv_sec - t[0].tv_sec)*1000.0 + (t[5].tv_usec - t[0].tv_usec)/1000.0;
  printf("Total time: %3.3f ms\n", total);  
  char *names[5] = {"Create", "Clean ", "Splat ", "Blur  ", "Slice "};
  
  for (int i = 1; i < 6; i++) {
    printf("%s: %3.3f ms\n", names[i-1], (t[i].tv_sec - t[i-1].tv_sec)*1000.0 + (t[i].tv_usec - t[i-1].tv_usec)/1000.0);
  }
  printf("Total GPU memory usage: %u bytes\n", (unsigned int)GPU_MEMORY_ALLOCATION);
  outputs.deviceToHost();
  destroyHashTable(table);
}

#ifdef LIBRARY
extern "C"
#ifdef WIN32
__declspec(dllexport)
#endif
#endif

/***************************************************/
/* A wrapper for the templatized version.          */
/***************************************************/
void filter(float *im, float *ref, float *slice, float *output, int pd, int vd, int imsize, int slicesize, bool accurate) {
  switch (vd*1000 + pd) {
  case 1003: filter_<1, 3>(im, ref, slice, output, imsize, slicesize, accurate); break;
  case 1006: filter_<1, 6>(im, ref, slice, output, imsize, slicesize, accurate); break;
  default:
    printf("Unsupported channel counts. Reference image must have 1 to 16 channels, input image must have 1 to 3 channels\n");	    
  }    
}

/***************************************************/
/* Command line wrapper                            */
/***************************************************/
#ifndef LIBRARY

struct header {
    int frames, width, height, channels;
};

void loadFLOAT(const char *filename, float **data, header *h) {
    FILE *f = fopen(filename, "rb");
    fread(h, sizeof(header), 1, f);
    size_t size = h->frames*h->width*h->channels*h->height;
    *data = new float[size];
    fread(*data, sizeof(float), size, f);
    fclose(f);
}

void saveFLOAT(const char *filename, float *data, header h) {
    FILE *f = fopen(filename, "wb");
    fwrite(&h, sizeof(header), 1, f);
    size_t size = h.frames*h.width*h.channels*h.height;
    fwrite(data, sizeof(float), size, f);
    fclose(f);
}

/*
 * Command-line interface for running the algorithm.
 *
 * It is assumed that the input files are in a particular format (.tmp) described in loadFLOAT and saveFLOAT routines above.
 * To be precise, the file begins with a header describing the image size (see the struct header), followed by a floating-point
 * array of size [width*height*channels*frame]. The floats are ordered by frame first, then row, then column, then channel.
 * The input files should have uniform width and height, have a single frame containing 5 channels (r,g,b,t,depth),
 * where depth of 0 indicates no data.
 * The output will have n frames of the same image size, containing two channels. The first channel is depth map; the second is
 * a confidence map.
 *
 * The standard deviations of the blurs are hard-coded in lines 76X. Change these to blur more or less in arbitrary dimensions.
 */
int main(int argc, char **argv) {
  initCuda(1, argv);
  srand(time(NULL));
  bool accurate = 1;

  // Check input parameters.
  if (argc < 3) {
    printf("Usage: %s [sampling percentage] output.tmp input1.tmp input2.tmp .... inputn.tmp\n", argv[0]);
    printf("  The sampling percentage specifies how many of the input depth points should be kept.\n");
    printf("  Set this to 100 for normal use. If you have a synthetic dataset with dense depthmap,\n");
    printf("  Lower this appropriately to avoid overwhelming the GPU memory.\n");
    return 1;
  }

  int nFrames = argc - 3;
  const short int MAX_PERFRAME = 20000; // max # of depth points per frame guaranteed by the user
  const short int FRAME_GROUP = 5; // This sets the size of neighborhood to blur.
  // 5 would use the preceding two frames, the current frame, and the next two frames.

  // Value vectors to filter. This is simply a 1-dimensional vector containing the depths.
  float * data_depth = new float[MAX_PERFRAME * FRAME_GROUP];
  // Position vectors to filter. This is a 6-dimensional vector containing r,g,b,x,y,t.
  float * data_position = new float[MAX_PERFRAME * FRAME_GROUP * 6];

  // The number of vectors in each frame
  unsigned int nData[FRAME_GROUP];
  unsigned int nDataSum = 0; // The total number of vectors over all frames being filtered collaboratively.
  float * sliceData[FRAME_GROUP]; // the position vectors to slice out in each frame.

  // Arrays to represent images
  float * output; // the values sliced for all frames
  float * file; // the floats read from a file.
  header outputHeader, fileHeader;

  // Standard deviations for filtering (their multiplicative inverses, in fact.)
  float INV_SPATIAL = 0.125f;
  float INV_TEMPORAL = 8.f;
  float INV_TONAL = 4.f;

  // Miscellaneous 
  timeval timer0, timer1, timer2;
  unsigned int offset = 0;
  float samplingPercentage = atof(argv[1]) * 0.01f;
  printf("Will keep %f percent of the input depth points.\n", samplingPercentage * 100);

  // Process each frame, using a moving window of width FRAME_GROUP.
  for (int i = 0; i < nFrames; i++) {   
    gettimeofday(&timer0, NULL);

    // Shift out an earlier frame that is out of the moving window.
    if (i >= FRAME_GROUP) {
      memcpy(data_depth, data_depth + nData[0], sizeof(float) * (nDataSum - nData[0]));
      memcpy(data_position, data_position + nData[0] * 6, sizeof(float) * (nDataSum - nData[0]) * 6);
      nDataSum -= nData[0];
      delete sliceData[0];
      for (int j = 0; j < FRAME_GROUP - 1; j++) {
	nData[j] = nData[j+1];
	sliceData[j] = sliceData[j+1];
      }
    }
    int index = (i < FRAME_GROUP) ? i : FRAME_GROUP - 1; // index for the next frame to go into the moving window

    // Read file
    loadFLOAT(argv[i+3], &file, &fileHeader);    

    // Allocate memory for the position vectors to slice out.
    sliceData[index] = new float[fileHeader.frames * fileHeader.width * fileHeader.height * 6];
    float * slice_ptr = sliceData[index];

    // Write current frame into the array for valid depth points.
    int nCount = 0;
    float * read_ptr = file;
    float * data_depth_write = data_depth + nDataSum;
    float * data_position_write = data_position + nDataSum * 6;
    for (int y = 0; y < fileHeader.height; y++) {
      for (int x = 0; x < fileHeader.width; x++) {
	if (read_ptr[4] > 0 && (float)rand() / (float)RAND_MAX < samplingPercentage) { // If the depth value is valid, include this point.	 
	  *data_depth_write = read_ptr[4];
	  data_depth_write++;
	  data_position_write[0] = (float)x * INV_SPATIAL;
	  data_position_write[1] = (float)y * INV_SPATIAL;
	  data_position_write[2] = read_ptr[0] * INV_TONAL;
	  data_position_write[3] = read_ptr[1] * INV_TONAL;
	  data_position_write[4] = read_ptr[2] * INV_TONAL;
	  data_position_write[5] = read_ptr[3] * INV_TEMPORAL;
	  data_position_write+=6;
	  nCount++;
	}
	slice_ptr[0] = (float)x * INV_SPATIAL;
	slice_ptr[1] = (float)y * INV_SPATIAL;
	slice_ptr[2] = read_ptr[0] * INV_TONAL;
	slice_ptr[3] = read_ptr[1] * INV_TONAL;
	slice_ptr[4] = read_ptr[2] * INV_TONAL;
	slice_ptr[5] = read_ptr[3] * INV_TEMPORAL;
	slice_ptr+=6;
	read_ptr+=5; // jump r,g,b,time,depth
      }
    }
    nData[index] = nCount;
    nDataSum += nCount;
    delete[] file;

    // Allocate memory for output once
    if (i == 0) {
      outputHeader.frames = nFrames - FRAME_GROUP + 1;
      outputHeader.channels = 2;
      outputHeader.width = fileHeader.width;
      outputHeader.height = fileHeader.height;
      output = new float[outputHeader.frames * outputHeader.channels *
			      outputHeader.width * outputHeader.height];
    }

    // Run the filter if enough frames are loaded to fill the moving window completely.
    gettimeofday(&timer1, NULL);
    if (i >= FRAME_GROUP - 1) {
      filter(data_depth, data_position, sliceData[FRAME_GROUP/2], output + offset,
	     6, 1, nDataSum, outputHeader.width*outputHeader.height, accurate);
      offset += outputHeader.width * outputHeader.height * 2;
    }
    printf("Frame %d: founds %d depth points, for a total of %d in the moving window.\n", i, nCount, nDataSum);
    gettimeofday(&timer2, NULL);

    printf("  >>>>>>>>> Frame processed in  %3.3f ms (filtering %3.3f ms / preprocess %3.3f ms)\n",
	   (timer2.tv_sec - timer0.tv_sec)*1000.0 + (timer2.tv_usec - timer0.tv_usec)/1000.0,
	   (timer2.tv_sec - timer1.tv_sec)*1000.0 + (timer2.tv_usec - timer1.tv_usec)/1000.0,
	   (timer1.tv_sec - timer0.tv_sec)*1000.0 + (timer1.tv_usec - timer0.tv_usec)/1000.0);
  }
  printf("Number of frames input : %d\n", nFrames);
  printf("Number of frames output: %d\n", outputHeader.frames);
  saveFLOAT(argv[2], output, outputHeader);
  for (int i = 0; i < FRAME_GROUP; i++) delete sliceData[i];
  delete[] data_depth;
  delete[] data_position;
  delete[] output;
  return 0;    
}

#endif
