/*************************************************************************/
/* Depth upsampling from sparse data                                     */
/*                                                                       */
/* For the paper, please see:                                            */
/*  Dolson, J., Baek, J., Plagmann, C. and Thrun, S.,                    */
/*   ``Upsampling Range Data in Dynamic Environments,'' In CVPR 2010.    */
/*                                                                       */
/* This is an implementation of a sparse hash table for the algorithm in */
/* the permutohedral-lattice filtering paper.                            */
/* In essence, a key with hash h is stored at the row-h of the hash      */
/* table, and if it is already occupied, it is stored at the next        */
/* available row.                                                        */
/* More accurately, the keys are in fact stored separately, and the hash */
/* table stores an index into the keys table.                            */
/*************************************************************************/

#include "cuda_memory.h"

/*************************************************************/
/* Flags for optimization                                    */
/*************************************************************/
/* There are two flags that can be turned on. For low dimensional cases,
 * we recommend turning off LINEAR_D_MEMORY, because it incurs a constant factor.
 */
#define USE_ADDITIVE_HASH // Exploit the additiveness of the hash function.
// #define LINEAR_D_MEMORY // Cache pointers back to the 0-color vertex for linear memory.

// Take n mod the size of hash table.
#define modHash(n) ((n)%(2*table->capacity));

struct HashTable {
    unsigned int *values;
    signed short *keys;
    int *entries;
    unsigned int *filled;
    unsigned int capacity;
    signed short * zero;
    char * rank;
};

// Cached hash of the offset vectors (permutations of {1,1,..,1,-d})
__device__ __constant__ unsigned int hOffset[64];

/*************************************************************/
/* Create an empty hash table                                */
/*************************************************************/
/* 
 * kd       : # of dimensions for the position vectors
 * vd       : # of dimensions for the value vectors
 * capacity : the maximal number of points in the hash table (this may grow.)
 */
template<int kd, int vd>
static HashTable *createHashTable(int capacity) {
    HashTable table;
    
    table.capacity = capacity;

    allocateCudaMemory((void**)&(table.values), capacity*vd*sizeof(unsigned int));
    CUDA_SAFE_CALL(cudaMemset((void *)table.values, 0, capacity*vd*sizeof(unsigned int)));

#ifndef LINEAR_D_MEMORY
    allocateCudaMemory((void **)&(table.keys), capacity*kd*sizeof(signed short));
    CUDA_SAFE_CALL(cudaMemset((void *)table.keys, 0, capacity*kd*sizeof(signed short)));
#endif

    allocateCudaMemory((void **)&(table.entries), capacity*2*sizeof(int));
    CUDA_SAFE_CALL(cudaMemset((void *)table.entries, -1, capacity*2*sizeof(int)));

    allocateCudaMemory((void **)&(table.filled), sizeof(int));
    CUDA_SAFE_CALL(cudaMemset((void *)table.filled, 0, sizeof(int)));

    HashTable *deviceTable;
    allocateCudaMemory((void **)&(deviceTable), sizeof(HashTable));
    CUDA_SAFE_CALL(cudaMemcpy(deviceTable, &table, sizeof(HashTable), cudaMemcpyHostToDevice));

    return deviceTable;   
}

/*************************************************************/
/* Reset the given hash table                                */
/*************************************************************/
template <int vd> static void resetHashTable(HashTable *t) {
  CUDA_SAFE_CALL(cudaMemset((void*)t->values, 0, t->capacity*vd*sizeof(unsigned int)));
}

/*************************************************************/
/* Release all memory associated with a given hash table.    */
/*************************************************************/
static void destroyHashTable(HashTable *t) {
    HashTable table;
    CUDA_SAFE_CALL(cudaMemcpy(&table, t, sizeof(HashTable), cudaMemcpyDeviceToHost));
    releaseCudaMemory(table.values);
    releaseCudaMemory(table.keys);
    releaseCudaMemory(table.entries);
    releaseCudaMemory(table.filled);
    releaseCudaMemory(t);
}

/*************************************************************/
/* Calculate the hash for the given key.                     */
/* This function is additive, and this fact can be exploited.*/
/*************************************************************/
/* 
 * There are two notions of hash in this file. One is a long (or full) hash
 * that is a 32-bit unsigned integer, and the other is a short hash, which is
 * the long hash modulo the capacity of the table.
 */
template<int kd> __device__ __host__ static unsigned int hash(HashTable *table, signed short *key) {
    unsigned int k = 0; 
    for (int i = 0; i < kd; i++) {
	k += key[i];
	k = k * 2531011; 
    }
    return k;
}

/*************************************************************/
/* The same function as above.                               */
/*************************************************************/
template<int kd> __device__ __host__ static unsigned int hash(HashTable *table, int *key) {
    unsigned int k = 0; 
    for (int i = 0; i < kd; i++) {
	k += key[i];
	k = k * 2531011; 
    }
    return k;
}

/*************************************************************/
/* Check if the key with the given 0-color neighbor and      */
/* a rank array equals the provided key.                     */
/*************************************************************/
/* 
 * zero : the pointer to the key of the zero-color neighbor
 * rank : the rank of the components of the offset vector.
 * idx  : the index of the key to be compared.
 * key  : a secondary key to be compared to the one being calculated.
 */
template<int d> __device__ static bool matchKey(signed short * zero, char * rank,
						int idx, signed short * key) {
  bool match = true;
  int slot = idx/(d+1), color = idx-slot*(d+1);
  rank += slot * (d+1); zero += slot * (d+1);
  for (int i = 0; i < d && match; i++) {
    match = (key[i] == zero[i] + color - (rank[i] > d-color ? (d+1) : 0));
  }
  return match;
}

/*************************************************************/
/* Calculate a key from a 0-color neighbor and a rank array. */
/* Each vertex in the lattice has a unique 0-color neighbor, */
/* so it can be recovered from the 0-color neighbor and      */
/* some encoding of the offset vector.                       */
/*************************************************************/
/*
 * zero : the pointer to the key of the zero-color neighbor
 * rank : the rank of the components of the offset vector.
 * idx  : the index of the key to be calculated.
 * key  : the output array.
 */
template<int d> __device__ static void generateKey(signed short * zero, char * rank, int idx, signed short * key) {
  int slot = idx/(d+1), color = idx-slot*(d+1);
  rank += slot * (d+1); zero += slot * (d+1);
  for (int i = 0; i < d; i++) {
    key[i] = zero[i] + color - (rank[i] > d-color ? (d+1) : 0);
  }
}

/*************************************************************/
/* Reset the values corresponding to the keys.               */
/*************************************************************/
static unsigned int* swapHashTableValues(HashTable *table, unsigned int *newValues) {
    HashTable localTable;
    CUDA_SAFE_CALL(cudaMemcpy((void *)&localTable, (void *)table, sizeof(HashTable), cudaMemcpyDeviceToHost));
    unsigned int *old = localTable.values;
    localTable.values = newValues;
    CUDA_SAFE_CALL(cudaMemcpy((void *)table, (void *)&localTable, sizeof(HashTable), cudaMemcpyHostToDevice));        
    return old;
}


/*************************************************************/
/* Insert an entry into the hash table                       */
/*************************************************************/
/*
 * table : the hash table into which the key shall be inserted
 * key   : the position vector to be inserted
 * slot  :
 */
template<int kd>
__device__ static int hashTableInsert(HashTable *table, signed short *key, unsigned int slot) {
  unsigned int myHash = hash<kd>(table, key);
  return hashTableInsert<kd>(table, myHash, key, slot);
}

/*************************************************************/
/* Insert an entry into the hash table. The full hash of the */
/* key is provided as an argument.                           */
/*************************************************************/
/*
 * table : the hash table into which the key shall be inserted
 * fh    : the full hash of the key
 * key   : the position vector to be inserted
 * slot  : the index in which the key is stored in the key table.
 */
template<int kd>
__device__ static int hashTableInsert(HashTable *table, unsigned int fh,
				      signed short *key, unsigned int slot) {	
  int h = modHash(fh); // Calculate the short hash.
  while (true) {
    int *e = table->entries + h;
    
    // if the cell is locked, spin
    volatile int *lock = e;	
    while (*lock == -2);
    
    // check if the cell is empty
    if (*e == -1) {
      int old = atomicCAS((int *)e, -1, -2); // grab the lock on this cell
      if (old != -1) continue; // someone else beat me to it, bail
      
#ifndef LINEAR_D_MEMORY
      for (int i = 0; i < kd; i++) {
	table->keys[slot*kd+i] = key[i];
      }
#endif
      
      *e = slot; // this releases the lock
      return h;
    }
    
    // check if the cell has a matching key
#ifdef LINEAR_D_MEMORY
    if (matchKey<kd>(table->zero, table->rank, (*e), key)) return h;
#else
    bool match = true;
    for (int i = 0; i < kd && match; i++) {
      match = (table->keys[(*e)*kd+i] == key[i]);
    }
    if (match) return h;
#endif
    
    // increment the bucket with wraparound
    h++;
    if (h == table->capacity*2) h = 0;
  }
}

/*************************************************************/
/* Retrieve the index of the entry with the given key in the */
/* hash table. The hash of the given key is provided in the  */
/* arguments, up to modulus by the capacity of the table.    */
/*************************************************************/
/*
 * table : the hash table being searched
 * fh    : the hash of the key being provided
 * key   : the position vector being looked up
 * slot  : the index in which the key is stored in the key table.
 */
template<int kd> __device__ static
int hashTableRetrieveWithHash(HashTable *table, unsigned int fh, signed short *key) {
  int h = modHash(fh); // Calculate the short hash
  while (true) {
    int *e = table->entries + h; // Examine the entry with the given hash    
    if (*e == -1) return -1; // Invalid entry, exit
    
    // Match the key directly.
#ifdef LINEAR_D_MEMORY
    if (matchKey<kd>(table->zero, table->rank, (*e), key)) return *e;
#else
    bool match = true;
    for (int i = 0; i < kd && match; i++) {
      match = (table->keys[(*e)*kd+i] == key[i]);
    }
    if (match) return *e;
#endif

    // Look at the next entry in the hash table.    
    h++;
    if (h == table->capacity*2) h = 0;
  }
}

/*************************************************************/
/* Retrieve the index of the entry with the given key in the */
/* hash table.                                               */
/*************************************************************/
/*
 * table : the hash table being searched
 * key   : the position vector being looked up
 */
template<int kd>
__device__ static int hashTableRetrieve(HashTable *table, signed short *key) {
  int h = modHash(hash<kd>(table, key)); // Calculate hash
  while (true) {
    int *e = table->entries + h; // Examine the entry with the given hash.
    if (*e == -1) return -1; // Invalid entry, exit.

    // Match the key directly.
#ifdef LINEAR_D_MEMORY
    if (matchKey<kd>(table->zero, table->rank, (*e), key)) return *e;
#else
    bool match = true;
    for (int i = 0; i < kd && match; i++) {
      match = (table->keys[(*e)*kd+i] == key[i]);
    }
    if (match) return *e;
#endif
    
    // Look at the next entry in the hash table.
    h++;
    if (h == table->capacity*2) h = 0;
  }
}
